@@ -43,6 +43,7 @@
 #endif
 
 #include "librio500.h"
+#include "usbdrv.h"
 #include "libpsf.h"
 #include "libfon.h"
 
@@ -50,40 +51,47 @@
 #include <byteswap.h>
 #endif
 
-int init_communication (struct usbdevice *rio_dev)
-{
-   int ret;
-  rio_dev->fd = usb_init(NULL, 0x0841, 0x0001);
+static unsigned verboselevel = 0;
 
-  if (rio_dev->fd == -1) {
+struct usbdevice *init_communication ()
+{
+  int ret;
+  int intf = 0;
+  struct usbdevice *rio_dev;
+
+  rio_dev = usb_open(USB_VENDOR_DIAMOND, USB_PRODUCT_DIAMOND_RIO500USB, 5000);
+
+  if (!rio_dev) {
     printf("usb_init returned failure\n");
-    return 1;
+    return NULL;
+  }
+
+  if(usb_claiminterface(rio_dev, intf)) {
+	printf("usb_claiminterface returned an error!\n");
   }
 
   send_command (rio_dev, START_USB_COMM, 0x00, 0x00);
-  return 0;
+  return rio_dev;
 }
 
 void
 finish_communication (struct usbdevice *rio_dev)
 {
+  int intf = 0;
+
   send_command (rio_dev, 0x42, 0x00, 0x00);
   send_command (rio_dev, END_USB_COMM, 0x00, 0x00);
   send_command (rio_dev, 0x42, 0x00, 0x00);
 
-  rio_usb_close(rio_dev);
+  if(usb_releaseinterface(rio_dev, intf)) {
+	printf("usb_releaseinterface returned an error!\n");
+  }
+  
+  usb_close(rio_dev);
 }
 
 void
-rio_usb_close(struct usbdevice *rio_usb)
-{
-/* Again we need to check out Thomas' usbdrvlinux and update for multiple
-   Rios ;) */
-   usb_close();
-}
-
-void
-send_folder_location (struct usbdevice *rio_dev, int offset, int folder_num)
+send_folder_location (struct usbdevice *rio_dev, int offset, int folder_num, int card)
 {
   folder_location location;
 
@@ -107,19 +115,22 @@ send_folder_location (struct usbdevice *rio_dev, int offset, int folder_num)
 }
 
 void
-format_flash (struct usbdevice *rio_dev)
+format_flash (struct usbdevice *rio_dev, int card)
 {
-  send_command (rio_dev, RIO_FORMAT_DEVICE, 0x2185, 0);
+  send_command (rio_dev, RIO_FORMAT_DEVICE, 0x2185, card);
   sleep (1); /* wait for flash memory to update */
 }
 
 mem_status *
-get_mem_status (struct usbdevice *rio_dev)
+get_mem_status (struct usbdevice *rio_dev, int card)
 {
   static mem_status status;
 
   memset (&status, 0, sizeof (mem_status));
 
+/* set card from which to get memory status */
+  send_command (rio_dev, 0x51, 1, card);
+
   rio_ctl_msg (rio_dev, RIO_DIR_IN, 0x57, 0, 0, sizeof(status), (void*)&status);
 
 /* this struct "filled" by the rio.  Need to switch to big_endian for ppc 
@@ -139,55 +150,90 @@ get_mem_status (struct usbdevice *rio_dev)
 }
 
 unsigned long
-query_mem_left (struct usbdevice *rio_dev)
+query_card_count (struct usbdevice *rio_dev)
+{
+  return ( ( (send_command (rio_dev, 0x42, 0, 0) & 0x40000000) >> 30) + 1);
+}
+
+
+unsigned long
+query_mem_left (struct usbdevice *rio_dev, int card)
 {
   unsigned long mem_left;
   send_command (rio_dev, 0x42, 0, 0);
-  send_command (rio_dev, 0x42, 0, 0);
-  mem_left = send_command (rio_dev, 0x50, 0, 0);
+  mem_left = send_command (rio_dev, 0x50, 0, card);
   /* wait a bit */
   send_command (rio_dev, 0x42, 0, 0);
-  send_command (rio_dev, 0x42, 0, 0);
 
   return mem_left;
 }
 
+unsigned long
+query_firmware_rev (struct usbdevice *rio_dev)
+{
+  return send_command (rio_dev, 0x40, 0, 0) & 0xffff;
+}
+
+unsigned long
+get_num_folder_blocks (struct usbdevice *rio_dev, int address, int card)
+{
+  unsigned long read_status=0;
+
+  read_status = send_command (rio_dev, 0x59, address, card);
+  /* if command fails end comm gracefully and return -1 */
+  if (read_status == 0)
+  {
+       //finish_communication(rio_dev);
+       send_command (rio_dev, 0x42, 0x00, 0x00);
+       send_command (rio_dev, END_USB_COMM, 0x00, 0x00);
+       send_command (rio_dev, 0x42, 0x00, 0x00);
+
+       return -1; 
+  }
+
+  return read_status;
+}
 
 GList *
-read_folder_entries (struct usbdevice *rio_dev)
+read_folder_entries (struct usbdevice *rio_dev, int card)
 {
   BYTE           *folder_block, *pb;
   GList          *entry_list = NULL;
   folder_entry   *entry, *copy;
   int            total_read, folder_count=0;
   unsigned long  com_status;
-	
-  /* Assign space for folder block */
-  folder_block = (BYTE *) malloc (FOLDER_BLOCK_SIZE);
+  unsigned long  folder_block_count;
+  unsigned long  total_folder_block_size;
+
+  /* Determine number of folder blocks */
+  folder_block_count = get_num_folder_blocks (rio_dev, 0xff00, card);
+  if (folder_block_count == -1)
+    return NULL;
+
+  total_folder_block_size = FOLDER_BLOCK_SIZE * folder_block_count;
+
+   /* Assign space for folder block */
+  folder_block = (BYTE *) malloc (total_folder_block_size);
   if (folder_block == NULL)
-    return NULL;
-
-  /* Read folder list */
-  /* This is a limitation. The rio allows you to have more than
-     one block for folders but i can't see how to get the amount
-     of folder blocks from the rio. For now i just read 1, 0x4000
-     byte block. Thats enough for 8 folders. */ 
-  com_status = send_read_command (rio_dev, 0xff00, 1);
-  if (com_status == -1)
-	return NULL;
-
-  total_read = rio_usb_bulk ( rio_dev, 0x81, folder_block, FOLDER_BLOCK_SIZE);
-
-  if (total_read != FOLDER_BLOCK_SIZE)
-    return NULL;
-
-  /* Make a list of entries: one for each folder. This code
-     will change when i learn how to read more than one folder
-     block. Right now the max num. folders is 8.
-   */
-  pb = folder_block;
-  entry = (folder_entry *)pb;
-  while (entry->offset != 0xffff && folder_count < 8)
+     return NULL;
+ 
+   /* Read folder list */
+  com_status = send_read_command (rio_dev, 0xff00, folder_block_count, card);
+  if (com_status == -1) {
+	free (folder_block);
+        return NULL;
+  }
+ 
+  total_read = bulk_read ( rio_dev, folder_block, total_folder_block_size);
+  if (total_read != total_folder_block_size) {
+     free (folder_block);
+     return NULL;
+  }
+ 
+  /* Make a list of entries: one for each folder. */
+   pb = folder_block;
+   entry = (folder_entry *)pb;
+  while (entry->offset != 0xffff && folder_count<8*folder_block_count)
   {
     copy = calloc (sizeof (folder_entry), 1);
     memcpy (copy, entry, sizeof (folder_entry));
@@ -206,12 +252,13 @@ read_folder_entries (struct usbdevice *rio_dev)
 
 
 GList *
-read_song_entries (struct usbdevice *rio_dev, GList *folder_entries, int folder_num)
+read_song_entries (struct usbdevice *rio_dev, GList *folder_entries, int folder_num, int card)
 {
   folder_entry *folder;
   song_entry *song, *song_copy;
   BYTE *song_block, *ps;
   GList *item, *song_list = NULL;
+  unsigned long com_status;
   int    total_read, address, num_blocks;
   int    size, count;
 
@@ -236,7 +283,7 @@ read_song_entries (struct usbdevice *rio_dev, GList *folder_entries, int folder_
   address = folder_num;
   address <<= 8;
   address |= 0x00ff;
-  address &= 0x0fff;
+  address &= 0xffff;
 
   size = num_blocks * FOLDER_BLOCK_SIZE;
   song_block = (BYTE *) malloc ( size+1 );
@@ -244,11 +291,17 @@ read_song_entries (struct usbdevice *rio_dev, GList *folder_entries, int folder_
     return NULL;
 
   /* Read folder list */
-  send_read_command (rio_dev, address, num_blocks);
+  com_status = send_read_command (rio_dev, address, num_blocks, card);
+  if (com_status == -1) {
+        free(song_block);
+        return NULL;
+   }
 
-  total_read = rio_usb_bulk (rio_dev, 0x81, song_block, size);
-  if ( total_read != size )
+total_read = rio_usb_bulk (rio_dev, 0x81, song_block, size);
+  if ( total_read != size ) {
+    free (song_block);
     return (NULL);
+  }
 
   /* Make list */
   ps = song_block;
@@ -275,7 +328,7 @@ read_song_entries (struct usbdevice *rio_dev, GList *folder_entries, int folder_
 }
 
 void
-write_folder_entries (struct usbdevice *rio_dev, GList *folder_list)
+write_folder_entries (struct usbdevice *rio_dev, GList *folder_list, int card)
 {
   int          num_blocks, list_len;
   int          count;
@@ -283,11 +336,12 @@ write_folder_entries (struct usbdevice *rio_dev, GList *folder_list)
   GList        *item;
   folder_entry *entry;
 
+  block = new_empty_block ();
+
   /* If there are no entries just send a blank block */
   if (folder_list == NULL)
   {
-    block = new_empty_block ();
-    send_write_command (rio_dev, 0xff00, 1);
+    send_write_command (rio_dev, 0xff00, 1, card);
     rio_usb_bulk (rio_dev, 2, block, 0x4000);
     free (block);
     return;
@@ -299,8 +353,8 @@ write_folder_entries (struct usbdevice *rio_dev, GList *folder_list)
   if (list_len & 0x7)
     num_blocks++;
 
-  send_write_command (rio_dev, 0xff00, num_blocks);
-  p = block = new_empty_block ();
+  send_write_command (rio_dev, 0xff00, num_blocks, card);
+  p = block;
   count = 0;
   for (item = g_list_first (folder_list); item; item = item->next)
   {
@@ -318,6 +372,7 @@ write_folder_entries (struct usbdevice *rio_dev, GList *folder_list)
       rio_usb_bulk (rio_dev, 2, block, 0x4000);
       count = 0;
       clear_block (block);
+      p = block;
     }
   }
 
@@ -330,7 +385,7 @@ write_folder_entries (struct usbdevice *rio_dev, GList *folder_list)
 }
 
 void
-write_song_entries (struct usbdevice *rio_dev, int folder_num, GList *song_list)
+write_song_entries (struct usbdevice *rio_dev, int folder_num, GList *song_list, int card)
 {
   int          num_blocks, list_len;
   int          count, address;
@@ -342,15 +397,16 @@ write_song_entries (struct usbdevice *rio_dev, int folder_num, GList *song_list)
   address = folder_num;
   address <<= 8;
   address |= 0x00ff;
-  address &= 0x0fff;
+  address &= 0xffff;
 
   block = new_empty_block ();
 
   /* If there are no entries just send a black block */
   if (song_list == NULL)
   {
-    send_write_command (rio_dev, address, 1);
+    send_write_command (rio_dev, address, 1, card);
     rio_usb_bulk (rio_dev, 2, block, 0x4000);
+    free(block);
     return;
   }
 
@@ -361,7 +417,7 @@ write_song_entries (struct usbdevice *rio_dev, int folder_num, GList *song_list)
     num_blocks++;
 
   /* Set to what folder this items go */
-  send_write_command (rio_dev, address, num_blocks);
+  send_write_command (rio_dev, address, num_blocks, card);
 
   /* Now write the items to the bulk pipe in 0x4000 byte chunks*/
   p = block;
@@ -546,15 +602,15 @@ int rio_usb_bulk(struct usbdevice *rio_dev, int ep, void *block, int size)
   void *data;
   int ret;
 
-/* The call to Thomas Sailer's usbdrvlinux routines don't take a file descriptor , so
-   for right now, we'll have to be happy with one Rio attached at a time */
   do {
     len = size - transmitted;
     data = (unsigned char *)block + transmitted;
-    ret = usb_bulk(ep, len, data);
+
+/* The 5000 is a timeout value */
+    ret = usb_bulk_msg(rio_dev, ep, len, data, 5000);
     if (ret < 0) {
-      printf("rio_usb_bulk: usb_bulk returned in error\n");
-      return ret;
+      printf("rio_usb_bulk: usb_bulk returned error %x\n", ret);
+      return -ret;
     } else {
       transmitted += ret;
     }
@@ -571,7 +627,7 @@ int rio_usb_bulk(struct usbdevice *rio_dev, int ep, void *block, int size)
 
 
 unsigned long
-send_write_command (struct usbdevice *rio_dev, int address, int num_blocks)
+send_write_command (struct usbdevice *rio_dev, int address, int num_blocks, int card)
 {   
   int length = num_blocks * 0x4000;
   int num_big_reads, num_small_reads;
@@ -582,9 +638,9 @@ send_write_command (struct usbdevice *rio_dev, int address, int num_blocks)
 
   /* rio returns 0 on command failure for 0x4f and 0x46. . we return -1 */
 
-  write_status = send_command (rio_dev, 0x4c, address, 0x00);
+  write_status = send_command (rio_dev, 0x4c, address, card);
 
-  write_status = send_command (rio_dev, 0x4f, 0xffff, 0x00);
+  write_status = send_command (rio_dev, 0x4f, 0xffff, card);
   /* if command fails end comm gracefully and return -1 */
   if (write_status == 0)
   {
@@ -604,7 +660,7 @@ send_write_command (struct usbdevice *rio_dev, int address, int num_blocks)
 }
 
 unsigned long
-send_read_command (struct usbdevice *rio_dev, int address, int num_blocks)
+send_read_command (struct usbdevice *rio_dev, int address, int num_blocks, int card)
 {
   int length = num_blocks * 0x4000;
   int num_big_reads, num_small_reads;
@@ -615,15 +671,7 @@ send_read_command (struct usbdevice *rio_dev, int address, int num_blocks)
 
   /* rio returns 0 on failure . . we use -1 */
 
-  read_status = send_command (rio_dev, 0x59, address, 0x00);
-  /* if command fails end comm gracefully and return -1 */
-  if (read_status == 0)
-  {
-	finish_communication(rio_dev);
-	return -1; 
-  }
-
-  read_status = send_command (rio_dev, 0x4e, address, 0x00);
+  read_status = send_command (rio_dev, 0x4e, address, card);
   /* if command fails end comm gracefully and return -1 */
   if (read_status == 0)
   {
@@ -672,9 +720,8 @@ rio_ctl_msg (struct usbdevice *rio_dev, int direction, unsigned char request, un
     requesttype = USB_DIR_OUT |
         USB_TYPE_VENDOR | USB_RECIP_DEVICE;
 
-/* The call to Thomas Sailer's usbdrvlinux routines don't take a file descriptor , so
-   for right now, we'll have to be happy with one Rio attached at a time */
-  ret = usb_control(requesttype, request, value, index, length, data);
+/* The 5000 is a timeout value */
+  ret = usb_control_msg(rio_dev, requesttype, request, value, index, length, data, 5000);
   return (ret < 0) ? -1 : 0;
 }
 
@@ -837,3 +884,25 @@ char *safe_strcat(char *dest, const char *src, size_t maxlength)
     dest[dest_len + src_len] = 0;
     return dest;
 }
+
+int lprintf(unsigned vl, const char *format, ...)
+{
+        va_list ap;
+        int r;
+
+        if (vl > verboselevel)
+                return 0;
+        va_start(ap, format);
+#ifdef HAVE_VSYSLOG
+        if (syslogmsg) {
+                static const int logprio[] = { LOG_ERR, LOG_INFO };
+                vsyslog((vl > 1) ? LOG_DEBUG : logprio[vl], format, ap);
+                r = 0;
+        } else
+#endif
+                r = vfprintf(stderr, format, ap);
+        va_end(ap);
+        return r;
+}
+
+
