@@ -41,7 +41,7 @@ finish_communication (int fd)
 }
 
 void
-send_folder_location (int fd, int offset, int folder_num)
+send_folder_location (int fd, int offset, int folder_num, int card)
 {
   folder_location location;
 
@@ -65,19 +65,22 @@ send_folder_location (int fd, int offset, int folder_num)
 }
 
 void
-format_flash (int fd)
+format_flash (int fd, int card)
 {
-  send_command (fd, RIO_FORMAT_DEVICE, 0x2185, 0);
+  send_command (fd, RIO_FORMAT_DEVICE, 0x2185, card);
   sleep (1); /* wait for flash memory to update */
 }
 
 mem_status *
-get_mem_status (int fd)
+get_mem_status (int fd, int card)
 {
   static mem_status status;
 
   memset (&status, 0, sizeof (mem_status));
 
+  /* set card from which to get memory status */
+  send_command (fd, 0x51, 1, card);
+
   rio_ctl_msg (fd, RIO_DIR_IN, 0x57, 0, 0, sizeof(status), (BYTE*)&status);
 
 /* this struct "filled" by the rio.  Need to switch to big_endian for ppc 
@@ -97,53 +100,88 @@ get_mem_status (int fd)
 }
 
 unsigned long
-query_mem_left (int fd)
+query_card_count (int fd)
+{
+  return ( ( (send_command (fd, 0x42, 0, 0) & 0x40000000) >> 30) + 1);
+}
+
+unsigned long
+query_mem_left (int fd, int card)
 {
   unsigned long mem_left;
-  mem_left = send_command (fd, 0x50, 0, 0);
-  /* wait a bit */
+
   send_command (fd, 0x42, 0, 0);
+  mem_left = send_command (fd, 0x50, 0, card);
+  /* wait a bit */
   send_command (fd, 0x42, 0, 0);
 
   return mem_left;
 }
 
+unsigned long
+query_firmware_rev (int fd)
+{
+  return send_command (fd, 0x40, 0, 0) & 0xffff;
+}
+
+unsigned long
+get_num_folder_blocks (int fd, int address, int card)
+{
+  unsigned long read_status=0;
+
+  read_status = send_command (fd, 0x59, address, card);
+  /* if command fails end comm gracefully and return -1 */
+  if (read_status == 0)
+  {
+       finish_communication(fd);
+       return -1; 
+  }
+
+  return read_status;
+}
+
 
 GList *
-read_folder_entries (int fd)
+read_folder_entries (int fd, int card)
 {
   BYTE           *folder_block, *pb;
   GList          *entry_list = NULL;
   folder_entry   *entry, *copy;
   int            total_read, folder_count=0;
   unsigned long  com_status;
-	
+  unsigned long  folder_block_count;
+  unsigned long  total_folder_block_size;
+
+  /* Determine number of folder blocks */
+  folder_block_count = get_num_folder_blocks (fd, 0xff00, card);
+  if (folder_block_count == -1)
+    return NULL;
+
+  total_folder_block_size = FOLDER_BLOCK_SIZE * folder_block_count;
+
   /* Assign space for folder block */
-  folder_block = (BYTE *) malloc (FOLDER_BLOCK_SIZE);
-  if (folder_block == NULL)
-    return NULL;
+  folder_block = (BYTE *) malloc (total_folder_block_size);
+   if (folder_block == NULL)
+     return NULL;
 
-  /* Read folder list */
-  /* This is a limitation. The rio allows you to have more than
-     one block for folders but i can't see how to get the amount
-     of folder blocks from the rio. For now i just read 1, 0x4000
-     byte block. Thats enough for 8 folders. */ 
-  com_status = send_read_command (fd, 0xff00, 1);
-  if (com_status == -1)
-	return NULL;
+   /* Read folder list */
+  com_status = send_read_command (fd, 0xff00, folder_block_count, card);
+   if (com_status == -1) {
+	free(folder_block);
+        return NULL;
+   }
 
-  total_read = bulk_read ( fd, folder_block, FOLDER_BLOCK_SIZE);
+  total_read = bulk_read ( fd, folder_block, total_folder_block_size);
 
-  if (total_read != FOLDER_BLOCK_SIZE)
-    return NULL;
+  if (total_read != total_folder_block_size) {
+     free(folder_block);
+     return NULL;
+  }
 
-  /* Make a list of entries: one for each folder. This code
-     will change when i learn how to read more than one folder
-     block. Right now the max num. folders is 8.
-   */
-  pb = folder_block;
-  entry = (folder_entry *)pb;
-  while (entry->offset != 0xffff && folder_count<8)
+  /* Make a list of entries: one for each folder. */
+   pb = folder_block;
+   entry = (folder_entry *)pb;
+  while (entry->offset != 0xffff && folder_count<8*folder_block_count)
   {
     copy = calloc (sizeof (folder_entry), 1);
     memcpy (copy, entry, sizeof (folder_entry));
@@ -162,12 +200,13 @@ read_folder_entries (int fd)
 
 
 GList *
-read_song_entries (int fd, GList *folder_entries, int folder_num)
+read_song_entries (int fd, GList *folder_entries, int folder_num, int card)
 {
   folder_entry *folder;
   song_entry *song, *song_copy;
   BYTE *song_block, *ps;
   GList *item, *song_list = NULL;
+  unsigned long  com_status;
   int    total_read, address, num_blocks;
   int    size, count;
 
@@ -200,11 +239,17 @@ read_song_entries (int fd, GList *folder_entries, int folder_num)
     return NULL;
 
   /* Read folder list */
-  send_read_command (fd, address, num_blocks);
+  com_status = send_read_command (fd, address, num_blocks, card);
+   if (com_status == -1) {
+	free(song_block);
+        return NULL;
+   }
 
   total_read = bulk_read (fd, song_block, size );
-  if ( total_read != size )
+  if ( total_read != size ) {
+    free(song_block);
     return (NULL);
+  }
 
   /* Make list */
   ps = song_block;
@@ -232,7 +277,7 @@ read_song_entries (int fd, GList *folder_entries, int folder_num)
 }
 
 void
-write_folder_entries (int fd, GList *folder_list)
+write_folder_entries (int fd, GList *folder_list, int card)
 {
   int          num_blocks, list_len;
   int          count;
@@ -240,11 +285,11 @@ write_folder_entries (int fd, GList *folder_list)
   GList        *item;
   folder_entry *entry;
 
+  block = new_empty_block();
   /* If there are no entries just send a blank block */
   if (folder_list == NULL)
   {
-    block = new_empty_block ();
-    send_write_command (fd, 0xff00, 1);
+    send_write_command (fd, 0xff00, 1, card);
     bulk_write (fd, block, 0x4000);
     free (block);
     return;
@@ -256,8 +301,8 @@ write_folder_entries (int fd, GList *folder_list)
   if (list_len & 0x7)
     num_blocks++;
 
-  send_write_command (fd, 0xff00, num_blocks);
-  p = block = new_empty_block ();
+  send_write_command (fd, 0xff00, num_blocks, card);
+  p = block;
   count = 0;
   for (item = g_list_first (folder_list); item; item = item->next)
   {
@@ -275,6 +320,7 @@ write_folder_entries (int fd, GList *folder_list)
       bulk_write (fd, block, 0x4000);
       count = 0;
       clear_block (block);
+      p = block;
     }
   }
 
@@ -287,7 +333,7 @@ write_folder_entries (int fd, GList *folder_list)
 }
 
 void
-write_song_entries (int fd, int folder_num, GList *song_list)
+write_song_entries (int fd, int folder_num, GList *song_list, int card)
 {
   int          num_blocks, list_len;
   int          count, address;
@@ -299,15 +345,16 @@ write_song_entries (int fd, int folder_num, GList *song_list)
   address = folder_num;
   address <<= 8;
   address |= 0x00ff;
-  address &= 0x0fff;
+  address &= 0xffff;
 
   block = new_empty_block ();
 
   /* If there are no entries just send a black block */
   if (song_list == NULL)
   {
-    send_write_command (fd, address, 1);
+    send_write_command (fd, address, 1, card);
     bulk_write (fd, block, 0x4000);
+    free(block);
     return;
   }
 
@@ -318,7 +365,7 @@ write_song_entries (int fd, int folder_num, GList *song_list)
     num_blocks++;
 
   /* Set to what folder this items go */
-  send_write_command (fd, address, num_blocks);
+  send_write_command (fd, address, num_blocks, card);
 
   /* Now write the items to the bulk pipe in 0x4000 byte chunks*/
   p = block;
@@ -546,7 +593,7 @@ bulk_read (int fd, BYTE *block, int num_bytes)
 
 
 unsigned long
-send_write_command (int fd, int address, int num_blocks)
+send_write_command (int fd, int address, int num_blocks, int card)
 {   
   int length = num_blocks * 0x4000;
   int num_big_reads, num_small_reads;
@@ -557,9 +604,9 @@ send_write_command (int fd, int address, int num_blocks)
 
   /* rio returns 0 on command failure for 0x4f and 0x46. . we return -1 */
 
-  write_status = send_command (fd, 0x4c, address, 0x00);
+  write_status = send_command (fd, 0x4c, address, card);
 
-  write_status = send_command (fd, 0x4f, 0xffff, 0x00);
+  write_status = send_command (fd, 0x4f, 0xffff, card);
   /* if command fails end comm gracefully and return -1 */
   if (write_status == 0)
   {
@@ -579,7 +626,7 @@ send_write_command (int fd, int address, int num_blocks)
 }
 
 unsigned long
-send_read_command (int fd, int address, int num_blocks)
+send_read_command (int fd, int address, int num_blocks, int card)
 {
   int length = num_blocks * 0x4000;
   int num_big_reads, num_small_reads;
@@ -588,17 +635,7 @@ send_read_command (int fd, int address, int num_blocks)
   num_big_reads   = length / 0x10000;
   num_small_reads = length % 0x10000;
 
-  /* rio returns 0 on failure . . we use -1 */
-
-  read_status = send_command (fd, 0x59, address, 0x00);
-  /* if command fails end comm gracefully and return -1 */
-  if (read_status == 0)
-  {
-	finish_communication(fd);
-	return -1; 
-  }
-
-  read_status = send_command (fd, 0x4e, address, 0x00);
+  read_status = send_command (fd, 0x4e, address, card);
   /* if command fails end comm gracefully and return -1 */
   if (read_status == 0)
   {
