@@ -1,6 +1,7 @@
 #include <librio500_api.h>
 
 #include <stdio.h>
+#include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
@@ -19,23 +20,23 @@ static int    file_size (char *filename);
 
 #ifndef WITH_USBDEVFS
 
-static int    remove_folder (int fd, int folder_num);
-static int    remove_song (int fd, int song_num, int folder_num);
-static int    is_first_folder (int fd);
+static int    remove_folder (int fd, int folder_num, int card);
+static int    remove_song (int fd, int song_num, int folder_num, int card);
+static int    is_first_folder (int fd, int card);
 static GList *add_song_to_list (GList *, char *fnm, int o, char *font, int fn);
-static void   add_folder (int fd, char *name, char *font_name, int font_number);
-static void   rename_folder (int rio_dev, int folder_num, char *name, char *font_name, int font_number);
-static void   rename_song (int rio_dev, int folder_num, int song_num, char *name, char *font_name, int font_number);
+static void   add_folder (int fd, char *name, char *font_name, int font_number, int card);
+static void   rename_folder (int rio_dev, int folder_num, char *name, char *font_name, int font_number, int card);
+static void   rename_song (int rio_dev, int folder_num, int song_num, char *name, char *font_name, int font_number, int card);
 
 #else /* With USBDEFVFS */
 
-static int    remove_folder (struct usbdevice *rio_dev, int folder_num);
-static int    remove_song (struct usbdevice *rio_dev, int song_num, int folder_num);
-static int    is_first_folder (struct usbdevice *rio_dev);
+static int    remove_folder (struct usbdevice *rio_dev, int folder_num, int card);
+static int    remove_song (struct usbdevice *rio_dev, int song_num, int folder_num, int card);
+static int    is_first_folder (struct usbdevice *rio_dev, int card);
 static GList *add_song_to_list (GList *, char *fnm, int o, char *font, int fn);
-static void   add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_number);
-static void rename_folder (struct usbdevice *rio_dev, int folder_num, char *name, char *font_name, int font_number);
-static void rename_song (struct usbdevice *rio_dev, int folder_num, int song_num, char *name, char *font_name, int font_number);
+static void   add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_number, int card);
+static void rename_folder (struct usbdevice *rio_dev, int folder_num, char *name, char *font_name, int font_number, int card);
+static void rename_song (struct usbdevice *rio_dev, int folder_num, int song_num, char *name, char *font_name, int font_number, int card);
 
 #endif /* WITH_USBDEVFS */
 
@@ -56,18 +57,25 @@ static gint   g_alpha_sort (gconstpointer a, gconstpointer b);
 
 /* -------------------------------------------------------------------
    NAME:        rio_new
-   DESCRIPTION: Create a new istance of the rio class.
+   DESCRIPTION: Create a new instance of the rio class.
    ------------------------------------------------------------------- */
 
 Rio500 *
 rio_new ()
 {
   Rio500  *instance;
+  char    *font_name = malloc(strlen(DEFAULT_FONT_PATH)+strlen(DEFAULT_FON_FONT)+1);
 
   instance = calloc (sizeof (Rio500), 1);
   if (instance == NULL)
     return NULL;
 
+  /* set defaults */
+  instance->font_num = 0;
+  strcpy(font_name,DEFAULT_FONT_PATH);
+  strcat(font_name, DEFAULT_FON_FONT);
+  instance->font = font_name;
+  instance->card = 0;
   rio_api_open_l (instance);
 
   return instance;
@@ -75,7 +83,7 @@ rio_new ()
 
 /* -------------------------------------------------------------------
    NAME:        rio_delete
-   DESCRIPTION: Destory instance of rio object.
+   DESCRIPTION: Destroy instance of rio object.
    ------------------------------------------------------------------- */
 
 void
@@ -124,7 +132,7 @@ rio_format (Rio500 *rio)
 
   start_comm (rio);
 
-  format_flash (rio->rio_dev);
+  format_flash (rio->rio_dev,rio->card);
 
   end_comm (rio);
 
@@ -147,6 +155,7 @@ rio_format (Rio500 *rio)
    {
      g_return_val_if_fail (rio, -1);
      rio->stat_func = func;
+     return TRUE;
    }
 
 
@@ -175,7 +184,7 @@ rio_get_content (Rio500 *rio)
   start_comm (rio);
 
   /* Read the stuff */
-  folders = read_folder_entries (rio->rio_dev);
+  folders = read_folder_entries (rio->rio_dev, rio->card);
   folder_num = 0;
   content = NULL;
 
@@ -236,7 +245,7 @@ rio_get_song_list (Rio500 *rio, GList *folders, int folder_num)
   start_comm (rio);
 
   /* Read the stuff */
-  songs = read_song_entries (rio->rio_dev, folders, folder_num);
+  songs = read_song_entries (rio->rio_dev, folders, folder_num, rio->card);
 
   /* Finish communication */
   end_comm (rio);
@@ -258,11 +267,10 @@ rio_add_song (Rio500 *rio, int folder_num, char *filename)
   int               font_number, mem_left;
   GList            *folders, *songs;
   folder_entry     *f_entry;
-  string            font_name = DEFAULT_FONT_PATH;
+  char             *font_name = rio->font;
 
   /* set defaults */
-  font_number = 0;
-  safe_strcat(font_name, DEFAULT_FON_FONT, sizeof(font_name));
+  font_number = rio->font_num;
 
   /* Init communication with rio */
   start_comm (rio);
@@ -272,9 +280,9 @@ rio_add_song (Rio500 *rio, int folder_num, char *filename)
      the device. So... try 3 times and make sure that it really is
      returning 0. */
   retries = 3;
-  while (retries > 0)
+  while (retries-- > 0)
   {
-    mem_left = query_mem_left (rio->rio_dev);
+    mem_left = query_mem_left (rio->rio_dev,rio->card);
     send_command (rio->rio_dev, 0x42, 0, 0);
     if (mem_left > 0)
       break;
@@ -286,10 +294,10 @@ rio_add_song (Rio500 *rio, int folder_num, char *filename)
   }
 
   /* Read folder & song block */
-  folders = read_folder_entries (rio->rio_dev);
+  folders = read_folder_entries (rio->rio_dev,rio->card);
   if ( folder_num > g_list_length (folders)-1 )
     folder_num = 0;
-  songs   = read_song_entries ( rio->rio_dev, folders, folder_num);
+  songs   = read_song_entries ( rio->rio_dev, folders, folder_num, rio->card);
 
   /* Write the song to the Rio */
   song_location = write_song (rio, filename);
@@ -303,7 +311,7 @@ rio_add_song (Rio500 *rio, int folder_num, char *filename)
   }
 
   /* Write song block to the correct folder */
-  write_song_entries (rio->rio_dev, folder_num, songs );
+  write_song_entries (rio->rio_dev, folder_num, songs ,rio->card);
   send_command (rio->rio_dev, 0x42, 0, 0);
   send_command (rio->rio_dev, 0x42, 0, 0);
 
@@ -314,16 +322,16 @@ rio_add_song (Rio500 *rio, int folder_num, char *filename)
   f_entry->offset = song_block_offset;
   f_entry->fst_free_entry_off += 0x800;
 
-  write_folder_entries ( rio->rio_dev, folders );
+  write_folder_entries ( rio->rio_dev, folders ,rio->card);
   send_command (rio->rio_dev, 0x42, 0, 0);
   send_command (rio->rio_dev, 0x42, 0, 0);
   folder_block_offset = send_command (rio->rio_dev, 0x43, 0x0, 0x0);
 
   /* Tell Rio where the root folder block is. */
-  send_folder_location (rio->rio_dev, folder_block_offset, folder_num);
+  send_folder_location (rio->rio_dev, folder_block_offset, folder_num, rio->card);
 
   /* Not really sure what this does */
-  send_command (rio->rio_dev, 0x58, 0x0, 0x0);
+  send_command (rio->rio_dev, 0x58, 0x0, rio->card);
 
   /* Close device */
   end_comm (rio);
@@ -342,7 +350,12 @@ rio_del_song (Rio500 *rio, int folder_num, int song_num)
 {
   int status = -1;
   g_return_val_if_fail (rio != NULL, -1);
+
+#ifndef WITH_USBDEVFS
   g_return_val_if_fail (rio->rio_dev > 0, -1);
+#else
+  g_return_val_if_fail (rio->rio_dev == NULL, -1);
+#endif
 
   /* Open connection to rio */
   start_comm (rio);
@@ -352,7 +365,7 @@ rio_del_song (Rio500 *rio, int folder_num, int song_num)
 	  return status;
 
 	/* Call internal function to delete song. */
-	status = remove_song (rio->rio_dev, song_num, folder_num);
+	status = remove_song (rio->rio_dev, song_num, folder_num, rio->card);
  
   /* Finish communication */
   end_comm (rio);
@@ -371,7 +384,12 @@ rio_del_folder (Rio500 *rio, int folder_num)
   int status = -1;
 
   g_return_val_if_fail (rio != NULL, -1);
+
+#ifndef WITH_USBDEVFS
   g_return_val_if_fail (rio->rio_dev > 0, -1);
+#else
+  g_return_val_if_fail (rio->rio_dev == NULL, -1);
+#endif
 
   /* Open connection to rio */
   start_comm (rio);
@@ -381,7 +399,7 @@ rio_del_folder (Rio500 *rio, int folder_num)
 	  return status;
 
 	/* Call internal function to delete song. */
-	status = remove_folder (rio->rio_dev, folder_num);
+	status = remove_folder (rio->rio_dev, folder_num, rio->card);
  
   /* Finish communication */
   end_comm (rio);
@@ -397,8 +415,8 @@ rio_del_folder (Rio500 *rio, int folder_num)
 int
 rio_add_folder (Rio500 *rio, char *folder_name)
 {
-  int font_number=0;
-  string font_name = DEFAULT_FONT_PATH;
+  int font_number=rio->font_num;
+  char *font_name = rio->font;
 #ifndef WITH_USBDEVFS
   int rio_dev;
 #else
@@ -408,12 +426,11 @@ rio_add_folder (Rio500 *rio, char *folder_name)
   g_return_val_if_fail (rio != NULL, -1);
   g_return_val_if_fail (folder_name != NULL, -1);
 
-  safe_strcat(font_name, DEFAULT_FON_FONT, sizeof(font_name));
 
   /* Init communication with rio */
   start_comm (rio);
 
-  add_folder (rio->rio_dev, folder_name, font_name, font_number);
+  add_folder (rio->rio_dev, folder_name, font_name, font_number, rio->card);
 
   /* Close device */
   end_comm (rio);
@@ -431,9 +448,9 @@ rio_add_folder (Rio500 *rio, char *folder_name)
 int
 rio_add_directory(Rio500 *rio, char *dir_name, int folder_num)
 {
-  int count=0, font_number=0, ret, mem_left, dirsize=0;
+  int count=0, font_number=rio->font_num, ret, mem_left, dirsize=0;
   int retries=0;
-  string font_name = DEFAULT_FONT_PATH;
+  char *font_name = rio->font;
   char *sp, *olddir;
   DIR *dp;
   struct dirent *de;
@@ -448,8 +465,6 @@ rio_add_directory(Rio500 *rio, char *dir_name, int folder_num)
   g_return_val_if_fail (rio != NULL, -1);
   g_return_val_if_fail (dir_name != NULL, -1);
 
-  safe_strcat(font_name, DEFAULT_FON_FONT, sizeof(font_name));
-
   /* Open up directory and find mp3's */
   
   dp = opendir(dir_name);
@@ -484,9 +499,9 @@ rio_add_directory(Rio500 *rio, char *dir_name, int folder_num)
      returning 0. */
 
   retries = 3;
-  while (retries > 0)
+  while (retries-- > 0)
   {
-    mem_left = query_mem_left (rio->rio_dev);
+    mem_left = query_mem_left (rio->rio_dev,0);
     send_command (rio->rio_dev, 0x42, 0, 0);
     if (mem_left > 0)
       break;
@@ -526,17 +541,20 @@ int
 rio_rename_folder (Rio500 *rio, int folder_num, char *folder_name)
 {
   int status = -1;
-  string font_name = DEFAULT_FONT_PATH;
-
+  char *font_name = rio->font;
+  int fnum = folder_num;
   g_return_val_if_fail (rio != NULL, -1);
-  g_return_val_if_fail (rio->rio_dev > 0, -1);
 
-  safe_strcat(font_name, DEFAULT_FON_FONT, sizeof(font_name));
+#ifndef WITH_USBDEVFS
+  g_return_val_if_fail (rio->rio_dev > 0, -1);
+#else
+  g_return_val_if_fail (rio->rio_dev == NULL, -1);
+#endif
 
   /* Open connection to rio */
   start_comm (rio);
 
-  rename_folder (rio->rio_dev, folder_num, folder_name, font_name, 0);
+  rename_folder (rio->rio_dev, fnum, folder_name, font_name, rio->font_num, rio->card);
 
   /* Finish communication */
   end_comm (rio);
@@ -553,17 +571,21 @@ int
 rio_rename_song (Rio500 *rio, int folder_num, int song_num, char *song_name)
 {
   int status = -1;
-  string font_name = DEFAULT_FONT_PATH;
+  char *font_name = rio->font;
+  int fnum = folder_num;
 
   g_return_val_if_fail (rio != NULL, -1);
-  g_return_val_if_fail (rio->rio_dev > 0, -1);
 
-  safe_strcat(font_name, DEFAULT_FON_FONT, sizeof(font_name));
+#ifndef WITH_USBDEVFS
+  g_return_val_if_fail (rio->rio_dev > 0, -1);
+#else
+  g_return_val_if_fail (rio->rio_dev == NULL, -1);
+#endif
 
   /* Open connection to rio */
   start_comm (rio);
 
-  rename_song (rio->rio_dev, folder_num, song_num, song_name, font_name, 0);
+  rename_song (rio->rio_dev, fnum, song_num, song_name, font_name, rio->font_num, rio->card);
 
   /* Finish communication */
   end_comm (rio);
@@ -572,7 +594,7 @@ rio_rename_song (Rio500 *rio, int folder_num, int song_num, char *song_name)
 }
 /* -------------------------------------------------------------------
    NAME:        rio_memory_left
-   DESCRIPTION: Return memory left in main memory.
+   DESCRIPTION: Return memory left in internal or external memory.
    ------------------------------------------------------------------- */
 
 unsigned long
@@ -589,13 +611,91 @@ rio_memory_left (Rio500 *rio)
 
   start_comm (rio);
 
-  result = query_mem_left (rio->rio_dev);
+  result = query_mem_left (rio->rio_dev,rio->card);
 
   end_comm (rio);
 
   return result;
 }
 
+/* -------------------------------------------------------------------
+   NAME:        rio_get_mem_total
+   DESCRIPTION: Return total memory available either internally
+		or externally.
+   ------------------------------------------------------------------- */
+
+unsigned long rio_get_mem_total(Rio500* rio)
+{
+    unsigned long result;
+    mem_status *status;
+
+    g_return_val_if_fail (rio != NULL, -1);
+#ifndef WITH_USBDEVFS
+    g_return_val_if_fail (rio->rio_dev > 0, -1);
+#else
+    g_return_val_if_fail (rio->rio_dev == NULL, -1);
+#endif
+
+    start_comm (rio);
+    status = get_mem_status(rio->rio_dev, rio->card);
+    result = (status->num_blocks * status->block_size);
+    end_comm (rio);
+
+    return result;
+}
+
+/* -------------------------------------------------------------------
+   NAME:        rio_set_font
+   DESCRIPTION: sets font name/number to be stored in rio500 struct.
+		By default this is filled with the compiled in font.
+		All fonts must be located in the default font path
+		and all must be either .fon fonts or .psf fonts
+   ------------------------------------------------------------------- */
+
+int
+rio_set_font (Rio500 *rio, char *font_name, int font_number)
+{
+  char *temp = DEFAULT_FONT_PATH;  
+  char *font = malloc(strlen(temp)+strlen(font_name)+1);
+ 
+  g_return_val_if_fail (rio != NULL, -1);
+#ifndef WITH_USBDEVFS
+  g_return_val_if_fail (rio->rio_dev > 0, -1);
+#else
+  g_return_val_if_fail (rio->rio_dev == NULL, -1);
+#endif
+
+  strcpy(font,DEFAULT_FONT_PATH);
+  strcat(font,font_name);
+
+  rio->font_num = font_number;
+  rio->font = font;
+  return 0;
+
+}
+
+/* -------------------------------------------------------------------
+   NAME:        rio_set_card
+   DESCRIPTION: Pass 1 if operations to involve external card
+                pass 0 if operations to use internal memory
+   ------------------------------------------------------------------- */
+
+int
+rio_set_card (Rio500 *rio, int card)
+{
+
+  g_return_val_if_fail (rio != NULL, -1);
+#ifndef WITH_USBDEVFS
+  g_return_val_if_fail (rio->rio_dev > 0, -1);
+#else
+  g_return_val_if_fail (rio->rio_dev == NULL, -1);
+#endif
+
+  rio->card = card;
+  return 0;
+
+}
+
 /* -------------------------------------------------------------------
 
                             Internal functions
@@ -619,7 +719,7 @@ rio_api_read_songs_l (Rio500 *rio, GList *folders, int num_folder)
   int song_num;
   RioSongEntry *song;
 
-  songs = read_song_entries (rio->rio_dev, folders, num_folder);
+  songs = read_song_entries (rio->rio_dev, folders, num_folder,rio->card);
   rio_song_list = NULL;
   song_num = 0;
   for (item = songs; item; item = item->next)
@@ -736,10 +836,10 @@ write_song (Rio500 *rio, char *filename)
   total = 0;
   blocks_left = num_blocks - num_chunks;
 
-  send_command (rio->rio_dev, 0x4f, 0xffff, 0);
+  send_command (rio->rio_dev, 0x4f, 0xffff, rio->card);
   while (blocks_left > 0)
   {
-    send_command (rio->rio_dev, 0x46, num_chunks, 00);
+    send_command (rio->rio_dev, 0x46, num_chunks, 0x0);
     for (j=0;j<num_chunks / 2;j++)
     {
       count = read (input_file, block, 0x20000);
@@ -826,9 +926,9 @@ file_size (char *filename)
 
 static int
 #ifndef WITH_USBDEVFS
-remove_folder (int rio_dev, int folder_num)
+remove_folder (int rio_dev, int folder_num, int card)
 #else
-remove_folder (struct usbdevice *rio_dev, int folder_num)
+remove_folder (struct usbdevice *rio_dev, int folder_num, int card)
 #endif
 {
   folder_entry     *folder;
@@ -836,40 +936,40 @@ remove_folder (struct usbdevice *rio_dev, int folder_num)
   GList            *folders, *songs;
 
   /* Read folder & song block */
-  folders = read_folder_entries (rio_dev);
+  folders = read_folder_entries (rio_dev,card);
   if ( folder_num > g_list_length (folders)-1 )
     return -1;
   else
-    songs   = read_song_entries (rio_dev, folders, folder_num); 
+    songs   = read_song_entries (rio_dev, folders, folder_num,card); 
 
 
   folder  = (folder_entry *)g_list_nth_data (g_list_first(folders), folder_num);
 
   for (song_num = g_list_length (songs) - 1; song_num >= 0; song_num--)
-    remove_song (rio_dev, song_num, folder_num);
+    remove_song (rio_dev, song_num, folder_num, card);
 
   folders = g_list_remove (folders, (gpointer) folder);
 
-  send_command (rio_dev, 0x4c, folder_num, 0xff);
-  write_folder_entries (rio_dev, folders);
+  send_command (rio_dev, 0x4c, ((folder_num << 8) | 0xff), card);
+  write_folder_entries (rio_dev, folders,card);
   send_command (rio_dev, 0x42, 0, 0);
   send_command (rio_dev, 0x42, 0, 0);
   folder_block_offset = send_command (rio_dev, 0x43, 0x0, 0x0);
 
   /* Tell Rio where the root folder block is. */
-  send_folder_location (rio_dev, folder_block_offset, 0);
+  send_folder_location (rio_dev, folder_block_offset, 0,card);
 
   /* Not really sure what this does */
-  send_command (rio_dev, 0x58, 0x0, 0x0);
+  send_command (rio_dev, 0x58, 0x0, card);
 
   return 0;
 }
 
 static int
 #ifndef WITH_USBDEVFS
-remove_song (int rio_dev, int song_num, int folder_num)
+remove_song (int rio_dev, int song_num, int folder_num, int card)
 #else
-remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
+remove_song (struct usbdevice *rio_dev, int song_num, int folder_num, int card)
 #endif
 {
   song_entry       *song;
@@ -878,15 +978,15 @@ remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
   GList            *folders, *songs;
 
    /* Read folder & song block */
-   folders = read_folder_entries (rio_dev);
+   folders = read_folder_entries (rio_dev,card);
    if (folders == NULL)
      return -1;
    if ( folder_num > g_list_length (folders)-1 )
    {
-     songs   = read_song_entries (rio_dev, folders, 0); /* use folder 0 by default */
+     songs   = read_song_entries (rio_dev, folders, 0,card); /* use folder 0 by default */
      folder_num = 0;
    } else
-     songs   = read_song_entries (rio_dev, folders, folder_num); 
+     songs   = read_song_entries (rio_dev, folders, folder_num,card); 
 
    /* Remove the song entry from the list */
    if ( songs == NULL || song_num > g_list_length (songs) -1)
@@ -897,10 +997,10 @@ remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
    songs = g_list_remove (songs, (gpointer)song);
 
    /* Send remove command to the rio */
-   send_command (rio_dev, 0x4c, ((folder_num << 8) | song_num), 0);
+   send_command (rio_dev, 0x4c, ((folder_num << 8) | song_num), card);
 
    /* Write song block to the correct folder */
-   write_song_entries (rio_dev, folder_num, songs );
+   write_song_entries (rio_dev, folder_num, songs, card);
    send_command (rio_dev, 0x42, 0, 0);
    send_command (rio_dev, 0x42, 0, 0);
 
@@ -911,16 +1011,16 @@ remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
    f_entry->offset = song_block_offset;
    f_entry->fst_free_entry_off -= 0x800;
 
-   write_folder_entries ( rio_dev, folders );
+   write_folder_entries ( rio_dev, folders, card);
    send_command (rio_dev, 0x42, 0, 0);
    send_command (rio_dev, 0x42, 0, 0);
    folder_block_offset = send_command (rio_dev, 0x43, 0x0, 0x0);
 
    /* Tell Rio where the root folder block is. */
-   send_folder_location (rio_dev, folder_block_offset, folder_num);
+   send_folder_location (rio_dev, folder_block_offset, folder_num,card);
 
    /* Not really sure what this does */
-   send_command (rio_dev, 0x58, 0x0, 0x0);
+   send_command (rio_dev, 0x58, 0x0, card);
 
    return 0;
 }
@@ -928,23 +1028,23 @@ remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
 
 static int
 #ifndef WITH_USBDEVFS
-is_first_folder (int rio_dev)
+is_first_folder (int rio_dev, int card)
 #else
-is_first_folder (struct usbdevice *rio_dev)
+is_first_folder (struct usbdevice *rio_dev, int card)
 #endif
 {
   int result;
 
-  result = send_command (rio_dev, 0x59, 0xff00, 0);
+  result = send_command (rio_dev, 0x59, 0xff00, card);
   if (result > 0)
     return FALSE;
 
   /* Try again just in case */
   send_command (rio_dev, 0x42, 0, 0);
-  result = send_command (rio_dev, 0x59, 0xff00, 0);
+  result = send_command (rio_dev, 0x59, 0xff00, card);
   send_command (rio_dev, 0x42, 0, 0);
   send_command (rio_dev, 0x42, 0, 0);
-  send_command (rio_dev, 0x58, 0x0, 0x0);
+  send_command (rio_dev, 0x58, 0x0, card);
   if (result > 0)
     return FALSE;
 
@@ -953,9 +1053,9 @@ is_first_folder (struct usbdevice *rio_dev)
 
 static void
 #ifndef WITH_USBDEVFS
-add_folder (int rio_dev, char *name, char *font_name, int font_number)
+add_folder (int rio_dev, char *name, char *font_name, int font_number, int card)
 #else
-add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_number)
+add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_number, int card)
 #endif
 {   
   GList *folders;
@@ -966,17 +1066,17 @@ add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_num
   folders = NULL;
 
   /* Check if this is the first folder */ 
-  if ( is_first_folder (rio_dev) )
+  if ( is_first_folder (rio_dev, card) )
   {
     folders = NULL; 
   } else {
-    folders = read_folder_entries (rio_dev);
+    folders = read_folder_entries (rio_dev,card);
   }
 
-  /* We can only have up to 8 folder entries: 0 - 7 */
+  /* We can only have up to 256 folder entries: 0 - 255 */
   last_folder = g_list_length (folders);
   
-  if ( last_folder > 7)
+  if ( last_folder > 255)
     return;
 
   /* Now create an new entry for the folder */
@@ -984,7 +1084,7 @@ add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_num
 
 
   /* Write song and folder blocks back to rio */
-  write_song_entries (rio_dev, last_folder, NULL);
+  write_song_entries (rio_dev, last_folder, NULL,card);
 
   /* Afer a write it is a good idea to wait a bit */
   send_command (rio_dev, 0x42, 0, 0);
@@ -997,7 +1097,7 @@ add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_num
   folders = g_list_append (folders, entry);
 
   /* Write folder list */
-  write_folder_entries (rio_dev, folders);
+  write_folder_entries (rio_dev, folders,card);
 
   /* Wait a bit after the read */
   send_command (rio_dev, 0x42, 0, 0);
@@ -1005,9 +1105,9 @@ add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_num
 
   /* Tell rio where the root folder block is */
   folder_block_loc = send_command (rio_dev, 0x43, 0, 0);
-  send_folder_location (rio_dev, folder_block_loc, last_folder);
+  send_folder_location (rio_dev, folder_block_loc, last_folder,card);
 
-  send_command (rio_dev, 0x58, 0x0, 0x0);
+  send_command (rio_dev, 0x58, 0x0, card);
 
   /* done */
   return;
@@ -1015,9 +1115,9 @@ add_folder (struct usbdevice *rio_dev, char *name, char *font_name, int font_num
 
 static void
 #ifndef WITH_USBDEVFS
-rename_folder (int rio_dev, int folder_num, char *name, char *font_name, int font_number)
+rename_folder (int rio_dev, int folder_num, char *name, char *font_name, int font_number, int card)
 #else
-rename_folder (struct usbdevice *rio_dev, int folder_num, char *name, char *font_name, int font_number)
+rename_folder (struct usbdevice *rio_dev, int folder_num, char *name, char *font_name, int font_number, int card)
 #endif
 {   
   GList *folders, *item;
@@ -1025,7 +1125,8 @@ rename_folder (struct usbdevice *rio_dev, int folder_num, char *name, char *font
   int   folder_block_loc;
   folder_entry *entry, *new_entry;
 
-  folders = read_folder_entries (rio_dev);
+  send_command(rio_dev,0x42,0,0);
+  folders = read_folder_entries (rio_dev,card);
 
   /* Check folder_num rang */
   last_folder = g_list_length (folders);
@@ -1048,7 +1149,7 @@ rename_folder (struct usbdevice *rio_dev, int folder_num, char *name, char *font
   }
 
   /* Write folder list */
-  write_folder_entries (rio_dev, folders);
+  write_folder_entries (rio_dev, folders,card);
 
   /* Wait a bit after the read */
   send_command (rio_dev, 0x42, 0, 0);
@@ -1056,9 +1157,9 @@ rename_folder (struct usbdevice *rio_dev, int folder_num, char *name, char *font
 
   /* Tell rio where the root folder block is */
   folder_block_loc = send_command (rio_dev, 0x43, 0, 0);
-  send_folder_location (rio_dev, folder_block_loc, last_folder);
+  send_folder_location (rio_dev, folder_block_loc, last_folder,card);
 
-  send_command (rio_dev, 0x58, 0x0, 0x0);
+  send_command (rio_dev, 0x58, 0x0, card);
 
   /* done */
   return;
@@ -1066,9 +1167,9 @@ rename_folder (struct usbdevice *rio_dev, int folder_num, char *name, char *font
 
 static void
 #ifndef WITH_USBDEVFS
-rename_song (int rio_dev, int folder_num, int song_num, char *name, char *font_name, int font_number)
+rename_song (int rio_dev, int folder_num, int song_num, char *name, char *font_name, int font_number, int card)
 #else
-rename_song (struct usbdevice *rio_dev, int folder_num, int song_num, char *name, char *font_name, int font_number)
+rename_song (struct usbdevice *rio_dev, int folder_num, int song_num, char *name, char *font_name, int font_number, int card)
 #endif
 {   
   GList *folders, *item, *songs;
@@ -1077,14 +1178,15 @@ rename_song (struct usbdevice *rio_dev, int folder_num, int song_num, char *name
   song_entry *entry, *new_entry;
   folder_entry *f_entry;
 
-  folders = read_folder_entries (rio_dev);
+  send_command(rio_dev,0x42,0,0);
+  folders = read_folder_entries (rio_dev,card);
 
   /* Check folder_num range */
   last_folder = g_list_length (folders);
   if (folder_num < 0 || folder_num > last_folder)
      folder_num = 0;
   
-  songs   = read_song_entries ( rio_dev, folders, folder_num);
+  songs   = read_song_entries ( rio_dev, folders, folder_num,card);
 
   /* Now create an new entry for the folder */
   new_entry = song_entry_new (name, font_name, font_number);
@@ -1106,7 +1208,7 @@ rename_song (struct usbdevice *rio_dev, int folder_num, int song_num, char *name
   }
 
   /* Write song block to the correct folder */
-  write_song_entries (rio_dev, folder_num, songs );
+  write_song_entries (rio_dev, folder_num, songs,card );
   send_command (rio_dev, 0x42, 0, 0);
   send_command (rio_dev, 0x42, 0, 0);
 
@@ -1116,19 +1218,19 @@ rename_song (struct usbdevice *rio_dev, int folder_num, int song_num, char *name
   f_entry = (folder_entry *) g_list_nth_data (folders, folder_num);
   f_entry->offset = song_block_offset;
 
-  write_folder_entries ( rio_dev, folders );
+  write_folder_entries ( rio_dev, folders,card);
   send_command (rio_dev, 0x42, 0, 0);
   send_command (rio_dev, 0x42, 0, 0);
   folder_block_offset = send_command (rio_dev, 0x43, 0x0, 0x0);
 
   /* Tell Rio where the root folder block is. */
-  send_folder_location (rio_dev, folder_block_offset, folder_num);
+  send_folder_location (rio_dev, folder_block_offset, folder_num,card);
 
    /* Not really sure what this does */
-  send_command (rio_dev, 0x58, 0x0, 0x0);
+  send_command (rio_dev, 0x58, 0x0, card);
 
   /* Write folder list */
-  write_folder_entries (rio_dev, folders);
+  write_folder_entries (rio_dev, folders,card);
 
   /* Wait a bit after the read */
   send_command (rio_dev, 0x42, 0, 0);
@@ -1136,9 +1238,9 @@ rename_song (struct usbdevice *rio_dev, int folder_num, int song_num, char *name
 
   /* Tell rio where the root folder block is */
   folder_block_offset = send_command (rio_dev, 0x43, 0, 0);
-  send_folder_location (rio_dev, folder_block_offset, last_folder);
+  send_folder_location (rio_dev, folder_block_offset, last_folder,card);
 
-  send_command (rio_dev, 0x58, 0x0, 0x0);
+  send_command (rio_dev, 0x58, 0x0, card);
 
   /* done */
   return;
@@ -1151,36 +1253,54 @@ start_comm (Rio500 *rio)
 {
   g_return_if_fail (rio != NULL);
 
+#ifndef WITH_USBDEVFS
   if (rio->rio_dev < 0)
     rio_api_open_l (rio);
+#else
+  if (rio->rio_dev == NULL)
+    rio->rio_dev = init_communication();
+#endif
+
 #ifndef WITH_USBDEVFS
   g_return_if_fail (rio->rio_dev > 0);
 #else
-  g_return_if_fail (rio->rio_dev == NULL);
+  g_return_if_fail (rio->rio_dev != NULL);
 #endif
 
   if (rio->stat_func)
     (*rio->stat_func) (0, "Opening rio device...",0);
 
+#ifndef WITH_USBDEVFS
   init_communication (rio->rio_dev);
+#endif
 }
 
 static void
 end_comm (Rio500 *rio)
 {
   g_return_if_fail (rio != NULL);
+#ifndef WITH_USBDEVFS
   if (rio->rio_dev < 0)
     rio_api_open_l (rio);
+#else
+  if (rio->rio_dev == NULL)
+    rio->rio_dev = init_communication();
+#endif
+
 #ifndef WITH_USBDEVFS
   g_return_if_fail (rio->rio_dev > 0);
 #else
-  g_return_if_fail (rio->rio_dev == NULL);
+  g_return_if_fail (rio->rio_dev != NULL);
 #endif
 
   if (rio->stat_func)
     (*rio->stat_func) (0, "Communication finished.",0);
 
   finish_communication (rio->rio_dev);
+
+#ifdef WITH_USBDEVFS
+  rio->rio_dev = NULL;
+#endif
 }
 
 gint g_alpha_sort (gconstpointer a, gconstpointer b)
