@@ -20,6 +20,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -28,9 +29,6 @@
 #include <time.h>
 #include <errno.h>
 #include "getopt.h"
-#ifdef USE_ID3_TAGS
-#include <string.h>
-#endif
 
 #include "librio500.h"
 
@@ -41,23 +39,34 @@
 
 #endif /* TRUE */
 
+#define STR_RIFF (((((('R' << 8) | 'I') << 8) | 'F') << 8) | 'F')
+#define STR_WAVE (((((('W' << 8) | 'A') << 8) | 'V') << 8) | 'E')
+#define STR_MPEG (((((('M' << 8) | 'P') << 8) | 'E') << 8) | 'G')
+#define STR_fmt  (((((('f' << 8) | 'm') << 8) | 't') << 8) | ' ')
+#define STR_fact (((((('f' << 8) | 'a') << 8) | 'c') << 8) | 't')
+#define STR_data (((((('d' << 8) | 'a') << 8) | 't') << 8) | 'a')
+
 void  usage (char *progname);
 #ifndef WITH_USBDEVFS
-int   write_song (int rio_dev, char *filename);
+int   write_song (int rio_dev, char *filename, int card_number);
 #else
-int   write_song (struct usbdevice *rio_dev, char *filename);
+int   write_song (struct usbdevice *rio_dev, char *filename, int card_number);
 #endif
 int   file_size (char *filename);
 char *strip_path (char *f);
 #ifdef USE_ID3_TAGS
 void
-get_some_switches (int argc, char *argv[], int *font_number, int *folder_num, char *display_format);
+get_some_switches (int argc, char *argv[], int *font_number, int *folder_num, char *display_format, int *card_number, int *card_auto);
 GList *add_song_to_list (GList *songs, char *filename, int offset, char *font_name, int font_number, char *display_format);
 #else
 void
-get_some_switches (int argc, char *argv[], int *font_number, int *folder_num);
+get_some_switches (int argc, char *argv[], int *font_number, int *folder_num, int *card_number, int *card_auto);
 GList *add_song_to_list (GList *songs, char *filename, int offset, char *font_name, int font_number);
 #endif
+static unsigned long get_frame_header(FILE *fp);
+static int is_frame_header(unsigned long fh);
+static int mp3_read_long(unsigned long *fhp, FILE *fp);
+static int mp3_read_byte(unsigned long *fhp, FILE *fp);
 
 char *font_name = DEFAULT_FONT_PATH;
 char *temp_name;
@@ -134,7 +143,8 @@ main(int argc, char *argv[])
 {
   int               retries, song_location, new_size;
   int               folder_block_offset, song_block_offset;
-  int               folder_num, font_number, mem_left;
+  int               folder_num, font_number, card_number, card_auto;
+  int 		    mem_left,filesize,card_changed;
   GList            *folders, *songs;
   mem_status       *mem;
   folder_entry     *f_entry;
@@ -154,15 +164,19 @@ main(int argc, char *argv[])
   /* set defaults, can be overridden with switches*/
   /* Tomoaki . . set the default value for display here (%7,%9) */
 
+  card_number = 0;
+  card_auto = 0;
   folder_num = 0;
   font_number = 0;
+  mem_left = 0;
+  folders = NULL;
 
 #ifdef USE_ID3_TAGS
   get_some_switches(argc,argv,&font_number,&folder_num,
-                    display_format);
+                    display_format,&card_number,&card_auto);
 
 #else
-  get_some_switches(argc,argv,&font_number,&folder_num);
+  get_some_switches(argc,argv,&font_number,&folder_num,&card_number,&card_auto);
 #endif
 
   /* quick fix (by Tomoaki) */
@@ -180,10 +194,6 @@ main(int argc, char *argv[])
         font_name=temp_name;
   }
 
-  while (optind < argc)  /* loop through filenames and add */
-  {
-        filename = argv[optind++];
-
   /* Open connection to rio */
 #ifndef WITH_USBDEVFS
   rio_dev = open (DEFAULT_DEV_PATH, O_RDWR);
@@ -197,7 +207,7 @@ main(int argc, char *argv[])
    /* Init communication with rio */
    init_communication (rio_dev);
 #else
-   if((init_communication(rio_dev)) != 0) {
+   if(!(rio_dev = init_communication())) {
      printf("init_communication() failed!\n");
      free(rio_dev);
      return -1;
@@ -205,41 +215,61 @@ main(int argc, char *argv[])
 #endif
 
    
-   /* Make sure there's enough space left */
+  while (optind < argc)  /* loop through filenames and add */
+  {
+        filename = argv[optind++];
+
+ /* Make sure there's enough space left */
    /* Sometimes quert_mem_left returns 0 but there really is space in
       the device. So... try 3 times and make sure that it really is
       returning 0. */
    retries = 3;
-   while (retries > 0)
+   while (retries-- > 0)
    {
-     mem_left = query_mem_left (rio_dev);
-     send_command (rio_dev, 0x42, 0, 0);
-     if (mem_left > 0)
-       break;
+     mem_left = query_mem_left (rio_dev,card_number);
+     send_command(rio_dev,0x42,0,0);
+     if (mem_left > 0) break;
    }
-   if (file_size (filename) > mem_left)
+     filesize = file_size(filename);
+     if (filesize > mem_left && card_auto && card_number==0) {
+        printf("\nNot enough room in internal memory\n");
+	printf("Autofill has been set\n");
+	printf("Trying external smartmedia card\n");
+	printf("Setting folder number = 0 on external card\n");
+	folder_num = 0;	
+	card_number++;
+	card_changed = 1;
+	mem_left = query_mem_left(rio_dev,card_number);
+     }
+   if (filesize > mem_left)
    {
-      printf ("Not enough space left in rio for this song.\n");
+      printf ("Not enough space left in rio for %s.\n",filename);
+      goto try_next;
       finish_communication (rio_dev);
-#ifdef WITH_USBDEVFS
-      free(rio_dev);
-#endif
       exit (0);
    }
    
-   /* Get the first free block and amont of free memory in Rio */
-   /* This doesn't seem to be necesary */
-   send_command (rio_dev, 0x50, 0, 0);
-   send_command (rio_dev, 0x51, 1, 0);
-   mem = get_mem_status (rio_dev);
+   mem_left -= filesize;
 
    /* Read folder & song block */
-   folders = read_folder_entries (rio_dev);
+   folders = read_folder_entries (rio_dev,card_number);
+
+   if (g_list_length(folders) == 0) {
+	if (card_number == 1) {
+        printf("At least one folder must be created on the smartmedia card before adding songs\n");
+	}
+	else {
+	printf("At least one folder must be created in internal memory before adding songs\n");
+        }
+	finish_communication (rio_dev);
+	exit(0);
+   }	
+
    if ( folder_num > g_list_length (folders)-1 )
      folder_num = 0;
-   songs   = read_song_entries ( rio_dev, folders, folder_num); 
+   songs   = read_song_entries ( rio_dev, folders, folder_num,card_number); 
    /* Write the song to the Rio */
-   song_location = write_song (rio_dev, filename);
+   song_location = write_song (rio_dev, filename,card_number);
    
    /* Add an entry to the song block */
 #ifdef USE_ID3_TAGS
@@ -254,14 +284,12 @@ main(int argc, char *argv[])
      finish_communication (rio_dev);
 #ifndef WITH_USBDEVFS
      close (rio_dev);
-#else
-     free(rio_dev);
 #endif
      exit (0);
    }
 
    /* Write song block to the correct folder */
-   write_song_entries (rio_dev, folder_num, songs );
+   write_song_entries (rio_dev, folder_num, songs,card_number );
    send_command (rio_dev, 0x42, 0, 0);
    send_command (rio_dev, 0x42, 0, 0);
 
@@ -281,7 +309,7 @@ main(int argc, char *argv[])
   because now the functions which read and write these entries do the
   byte swapping so the structures are endian correct (i hope).
 */
-   write_folder_entries ( rio_dev, folders );
+   write_folder_entries ( rio_dev, folders,card_number );
    send_command (rio_dev, 0x42, 0, 0);
    send_command (rio_dev, 0x42, 0, 0);
    folder_block_offset = send_command (rio_dev, 0x43, 0x0, 0x0);
@@ -291,19 +319,18 @@ main(int argc, char *argv[])
 #endif
 
    /* Tell Rio where the root folder block is. */
-   send_folder_location (rio_dev, folder_block_offset, folder_num);
+   send_folder_location (rio_dev, folder_block_offset, folder_num,card_number);
 
    /* Not really sure what this does */
-   send_command (rio_dev, 0x58, 0x0, 0x0);
+   send_command (rio_dev, 0x58, 0x0, card_number);
+
+try_next:
+   } /* end of add file loop */
 
    /* Close device */
    finish_communication (rio_dev);
 #ifndef WITH_USBDEVFS
    close (rio_dev);
-#endif
-   } 
-#ifdef WITH_USBDEVFS
-   free (rio_dev);
 #endif
    exit (0);
 }
@@ -341,7 +368,6 @@ add_song_to_list (GList *songs, char *filename, int offset, char *font_name, int
   fp = fopen (filename, "r");
   if (fp == NULL)
     return NULL;
-  fclose (fp);
 
   size = file_size (filename);
 
@@ -350,8 +376,9 @@ add_song_to_list (GList *songs, char *filename, int offset, char *font_name, int
   entry->offset = (WORD)  offset;
   entry->length = (DWORD) size;
   entry->dunno3 = (WORD) 0x0020;
-  entry->mp3sig = (DWORD) 0x0092fbff;
+  entry->mp3sig = (DWORD) get_frame_header(fp);
   entry->time   = (DWORD) time (NULL);
+  fclose (fp);
 
 #ifdef USE_ID3_TAGS
   if ((striped_name = get_display_string (display_format, filename)) == NULL)
@@ -382,6 +409,99 @@ add_song_to_list (GList *songs, char *filename, int offset, char *font_name, int
   return g_list_first (new_entry);
 }
 
+static unsigned long
+get_frame_header(FILE *fp)
+{
+	unsigned long fh = 0;
+	int i;
+
+	if (!mp3_read_long(&fh, fp)) 
+		return 0;
+	if (fh == STR_RIFF) {
+riff:		if (!mp3_read_long(&fh, fp)) 
+			return 0;
+		if (!mp3_read_long(&fh, fp)) 
+			return 0;
+		if (fh != STR_MPEG) {
+			if (fh != STR_WAVE) {
+				goto search;
+			}
+			if (!mp3_read_long(&fh, fp)) 
+				return 0;
+			if (fh != STR_fmt) {
+				goto search;
+			}
+			if (!mp3_read_long(&fh, fp)) 
+				return 0;
+			if (fh > 4096) {
+				goto search;
+			}
+			for (i = fh; i--; ) 
+				if (getc(fp) == EOF) 
+					return 0;
+		}
+		if (!mp3_read_long(&fh, fp)) 
+			return 0;
+		if (fh == STR_fact) {
+			if (!mp3_read_long(&fh, fp)) 
+				return 0;
+			if (fh > 4096) {
+				goto search;
+			}
+			for (i = fh; i--; ) 
+				if (getc(fp) == EOF) 
+					return 0;
+			if (!mp3_read_long(&fh, fp)) 
+				return 0;
+		}
+		if (fh != STR_data) {
+			goto search;
+		}
+		if (!mp3_read_long(&fh, fp)) 
+			return 0;
+		if (!mp3_read_long(&fh, fp)) 
+			return 0;
+	}
+
+search:	i = 65536;
+	while (!is_frame_header(fh)) {
+		if (fh == STR_RIFF) 
+			goto riff;
+		if (i-- == 0 || !mp3_read_byte(&fh, fp)) 
+			return 0;
+	}
+	return GUINT32_SWAP_LE_BE(fh);
+}
+
+static int
+is_frame_header(unsigned long fh)
+{
+	return
+		(fh & 0xffe00000) == 0xffe00000 &&
+		((fh >> 17) & 3) != 0 &&
+		((fh >> 12) & 0xf) != 0xf &&
+		((fh >> 10) & 0x3) != 0x3 &&
+		(fh & 0xffff0000) != 0xfffe0000;
+}
+
+static int
+mp3_read_long(unsigned long *fhp, FILE *fp)
+{
+	return
+		mp3_read_byte(fhp, fp) && mp3_read_byte(fhp, fp) &&
+		mp3_read_byte(fhp, fp) && mp3_read_byte(fhp, fp);
+}
+
+static int
+mp3_read_byte(unsigned long *fhp, FILE *fp)
+{
+	int n = getc(fp);
+
+	if (n == EOF) return FALSE;
+	*fhp = (*fhp << 8) | n;
+	return TRUE;
+}
+
 #ifdef USE_ID3_TAGS
 #define GET_ID3TAG( _TAGNAME )                                            \
   _TAGNAME = (char *) calloc( sizeof(id3tag_v1->_TAGNAME) +1, 1);         \
@@ -404,7 +524,7 @@ get_display_string(char *display_format, char *filename)
   char      *title, *artist, *album, *year, *comment;
   char      *ptr;
   int       count;
-
+  char 	    track[4]; 
   if ((id3tag_v1 = get_id3info_v1 (filename)) == NULL)
     return NULL;
 
@@ -426,7 +546,7 @@ get_display_string(char *display_format, char *filename)
       ext  = tmp_ptr+1;
       break;
     }
-    *tmp_ptr--;
+    tmp_ptr--;
   }
 /*
 printf("body=%s\next=%s\npath=%s\nfilename=%s\n", body,ext, path,filename);
@@ -439,6 +559,11 @@ printf("body=%s\next=%s\npath=%s\nfilename=%s\n", body,ext, path,filename);
   GET_ID3TAG(year);
   GET_ID3TAG(comment);
 
+/* id3v1.1: if comment[28] is 0, comment[29] has track number */
+   if(id3tag_v1->comment[28]=='\0')
+	sprintf(track, "%d", id3tag_v1->comment[29]);
+   else
+	track[0]=0;
 /*
 printf("artist: [%s]\ntitle:  [%s]\nalbum:  [%s]\nyear:   [%s]\ncomment:[%s]\ngenre:[%s]\n",+ artist, title, album, year, comment, mpg123_id3_genres[id3tag_v1->genre]);
 printf("%s\n", display_format);
@@ -452,6 +577,10 @@ printf("%s\n", display_format);
     if (*ptr == '%') {
       ptr++; count++;
       switch (*ptr) {
+        case '0':				       /* %0: v1.1 Track # */
+	 safe_strcat(display_string, track, DISPLAY_STRING_LEN);
+	 break;
+
         case '1':                                      /* %1: ID3 Artist */
          safe_strcat(display_string, artist, DISPLAY_STRING_LEN);
          break;
@@ -548,9 +677,9 @@ get_id3info_v1(char *filename)
   
 int
 #ifndef WITH_USBDEVFS
-write_song (int rio_dev, char *filename)
+write_song (int rio_dev, char *filename, int card)
 #else
-write_song (struct usbdevice *rio_dev, char *filename)
+write_song (struct usbdevice *rio_dev, char *filename, int card)
 #endif
 {
   int input_file;
@@ -579,10 +708,10 @@ write_song (struct usbdevice *rio_dev, char *filename)
   total = 0;
   blocks_left = num_blocks - num_chunks;
  
-  send_command (rio_dev, 0x4f, 0xffff, 0);
+  send_command (rio_dev, 0x4f, 0xffff, card);
   while (blocks_left > 0)
   {
-    send_command (rio_dev, 0x46, num_chunks, 00);
+    send_command (rio_dev, 0x46, num_chunks, 0);
     for (j=0;j<num_chunks / 2;j++)
     {
       count = read (input_file, block, 0x20000);
@@ -601,8 +730,8 @@ write_song (struct usbdevice *rio_dev, char *filename)
 
   /* Send remaining blocks */
   blocks_left += num_chunks;
-  //send_command (rio_dev, 0x4f, 0xffff, 0);
-  send_command (rio_dev, 0x46, blocks_left, 00);
+  //send_command (rio_dev, 0x4f, 0xffff, card);
+  send_command (rio_dev, 0x46, blocks_left, 0);
   while (blocks_left > 0)
   {
     count = read (input_file, block, 0x10000);
@@ -626,11 +755,11 @@ write_song (struct usbdevice *rio_dev, char *filename)
   {
     if (j > 0x4000)
     {
-      //send_command (rio_dev, 0x4f, 0xffff, 0);
+      //send_command (rio_dev, 0x4f, 0xffff, card);
       send_command (rio_dev, 0x46, 00, 0x4000);
       bulk_write (rio_dev, p, 0x4000);
     } else {
-      //send_command (rio_dev, 0x4f, 0xffff, 0);
+      //send_command (rio_dev, 0x4f, 0xffff, card);
       send_command (rio_dev, 0x46, 00, j);
       bulk_write (rio_dev, p, j);
     }
@@ -685,30 +814,29 @@ strip_path (char *f)
 int 
 file_size (char *filename)
 {
-  int input_file;
   struct stat file_stat;
   int  size;
 
-  input_file = open (filename, O_RDONLY);
-  if (input_file == -1)
+  if (stat(filename,&file_stat) == -1)
     return 0;
 
-  stat (filename, &file_stat);
   size = file_stat.st_size;
 
   return size;
 }
 
 #ifdef USE_ID3_TAGS
-static char const shortopts[] = "d:F:f:n:hv";
+static char const shortopts[] = "d:xaF:f:n:hv";
 #else
-static char const shortopts[] = "F:f:n:hv"; 
+static char const shortopts[] = "xaF:f:n:hv"; 
 #endif
 static struct option const longopts[] =
 {
 #ifdef USE_ID3_TAGS
   {"display", required_argument, NULL, 'd'},
 #endif
+  {"external", no_argument, NULL, 'x'},
+  {"autofill", no_argument, NULL, 'a'},
   {"folder", required_argument, NULL, 'F'},
   {"fontname", required_argument, NULL, 'f'},
   {"fontnumber", required_argument, NULL, 'n'},
@@ -728,10 +856,13 @@ static char const *const option_help[] =
 "                               %1=ID3 Artist  %2=ID3 Title    %3=ID3 Album",
 "                               %4=ID3 Year    %5=ID3 Comment  %6=ID3 Genre",
 "                               %7=File Name   %8=File Path    %9=File Extension",
+"                               %0=Track Number (id3v1.1)",
 "",
 "                               default FORMAT is %7.%9",
 "",
 #endif
+"  -x        --external         Write to external memory card",
+"  -a        --autofill         Use external memory card if internal full",
 "  -F x      --folder x         Transfer song(s) into folder of index=x",
 "  -f name   --fontname name    Set the fontname to be used on the Rio display.",
 "  -n x      --fontnumber x     Set the fontnumber within the given ",
@@ -751,11 +882,11 @@ static char const *const option_help[] =
 
 #ifdef USE_ID3_TAGS
 void
-get_some_switches (int argc, char *argv[], int *font_number, int *folder_num, char *display_format)
+get_some_switches (int argc, char *argv[], int *font_number, int *folder_num, char *display_format, int *card_number, int *card_auto)
 #else
 
 void
-get_some_switches (int argc, char *argv[], int *font_number, int *folder_num)
+get_some_switches (int argc, char *argv[], int *font_number, int *folder_num, int *card_number, int *card_auto)
 #endif
 {
     register int optc;
@@ -781,6 +912,15 @@ get_some_switches (int argc, char *argv[], int *font_number, int *folder_num)
 		break;
 #endif
 
+	    case 'x':
+		*card_number = 1;
+		break;
+	  
+	    case 'a':
+		*card_number = 0;
+		*card_auto = 1;
+		break;
+
 	    case 'F':
 		/* Sanity check --foldernumber digit */
 		if(!isdigit(*optarg)) {
