@@ -41,6 +41,7 @@
 int name_flag = 0;
 int whole_folder = 0;
 int folder_num_set = 0;
+int card_number = 0;
 char *folder_num_string = NULL;
 char numstring[4];
 
@@ -48,14 +49,14 @@ static gint   g_alpha_sort (gconstpointer a, gconstpointer b);
 char * int_to_string(int x);
 void usage (char *progname);
 int main(int argc, char *argv[]);
-void get_some_switches (int argc, char *argv[], int *folder_num, int *automat);
+void get_some_switches (int argc, char *argv[], int *folder_num, int *automat, int *card);
 
 #ifndef WITH_USBDEVFS
-int remove_folder (int rio_dev, int folder_num);
-int remove_song (int rio_dev, int song_num, int folder_num);
+int remove_folder (int rio_dev, int folder_num, int card_number);
+int remove_song (int rio_dev, int song_num, int folder_num, int card_number);
 #else
-int remove_folder (struct usbdevice *rio_dev, int folder_num);
-int remove_song (struct usbdevice *rio_dev, int song_num, int folder_num);
+int remove_folder (struct usbdevice *rio_dev, int folder_num, int card_number);
+int remove_song (struct usbdevice *rio_dev, int song_num, int folder_num, int card_number);
 #endif
 
 
@@ -74,7 +75,7 @@ int
 main(int argc, char *argv[])
 {
   int		    song_retries,folder_retries;
-  int		    folder_list_length,song_list_length;
+  int		    folder_list_length,song_list_length = 0;
   int               match_flag,count,counter,song_num,folder_num;
   int               num_things_to_delete;
   char		    *song_num_string=NULL;
@@ -109,7 +110,7 @@ main(int argc, char *argv[])
   folder_num=song_num=num_things_to_delete=0;
 
   /* Process switches */
-  get_some_switches(argc,argv,&folder_num,&automatic);
+  get_some_switches(argc,argv,&folder_num,&automatic,&card_number);
 
   /* Build folder and song lookup lists */
 
@@ -126,7 +127,7 @@ main(int argc, char *argv[])
    init_communication (rio_dev);
 #else
 
-   if((init_communication(rio_dev)) != 0) {
+   if(!(rio_dev = init_communication())) {
      printf("init_communication() failed!\n");
      return -1;
    }
@@ -136,7 +137,9 @@ main(int argc, char *argv[])
   /* sometimes read_folder_entries fails . . try 3 times */
   while (folder_list_length == 0 && folder_retries < 3)
   {
-    rio_folders = read_folder_entries (rio_dev);
+    send_command (rio_dev, 0x42, 0, 0);
+
+    rio_folders = read_folder_entries (rio_dev,card_number);
     folder_list_length = g_list_length(rio_folders);
     folder_retries++;
   }
@@ -163,7 +166,7 @@ main(int argc, char *argv[])
    
     while (song_list_length == 0 && song_retries < 3)
      {
-       song_lists[count]= read_song_entries (rio_dev, rio_folders, count);
+       song_lists[count]= read_song_entries (rio_dev, rio_folders, count, card_number);
        song_list_length = g_list_length(song_lists[count]);
        song_retries++;
      }
@@ -199,13 +202,10 @@ main(int argc, char *argv[])
    }
 
   /* Make sure folder_num isn't set to something stupid with -F */
-  if (folder_num < 0 || folder_num>7)
+  if (folder_num < 0 || folder_num>256) /* new limit is 256 folders */
   {
     printf ("Non-existent folder\n");
     usage (argv[0]);
-#ifdef WITH_USBDEVFS
-    free(rio_dev);
-#endif
     exit (-1);
   }
 
@@ -248,7 +248,7 @@ main(int argc, char *argv[])
       /* end of match loop, check that a match was found */
       if (match_flag == 0)
       {
-  	printf("\n%s did not match any folder stored on the rio500\n",arg_ent->data);
+  	printf("\n%s did not match any folder stored on the rio500\n",(char*)arg_ent->data);
 	printf("Check the folder names and try again\n");
 	exit(-1);
       }
@@ -278,7 +278,7 @@ main(int argc, char *argv[])
       /* end of match loop, check that a match was found */
       if (match_flag == 0)
       {
-        printf("\n%s did not match any song stored on the rio500\n",arg_ent->data);
+        printf("\n%s did not match any song stored on the rio500\n",(char*)arg_ent->data);
         printf("Check the folder names and try again\n");
         exit(-1);
 
@@ -303,13 +303,13 @@ main(int argc, char *argv[])
 	}
 
 	/* sanity check -- make sure folder num, song num exist on rio */
-	if (folder_num >= folder_list_length)
+	if ((whole_folder == 1) && (folder_num >= folder_list_length))
 	{
 	  printf("\nYou are attempting to delete a non-existent folder index");
 	  printf("\nCheck the folder index number and try again\n");
 	  exit(-1);
 	}
-	if (song_num >= song_list_length)
+	if ((whole_folder == 0) && (song_num >= g_list_length(song_lists[folder_num])))
 	{
           printf("\nYou are attempting to delete a non-existent song index");
 	  printf("\nCheck the song index number and try again\n");
@@ -322,9 +322,6 @@ main(int argc, char *argv[])
     {
       printf ("\nVerify that the rio500.o module is loaded, and your Rio is \n");
       printf ("connected and powered up.\n\n");
-#ifdef WITH_USBDEVFS
-    free(rio_dev);
-#endif
       exit (-1);
     }
 
@@ -332,7 +329,7 @@ main(int argc, char *argv[])
    init_communication (rio_dev);
 #else
 
-   if((init_communication(rio_dev)) != 0) {
+   if(!(rio_dev = init_communication())) {
      printf("init_communication() failed!\n");
      free(rio_dev);
      return -1;
@@ -353,13 +350,11 @@ main(int argc, char *argv[])
 	     finish_communication (rio_dev);
 #ifndef WITH_USBDEVFS
 	     close (rio_dev);
-#else
-             free(rio_dev);
 #endif
 	     exit (0);
 	   }
        }
-     remove_folder (rio_dev, folder_num);
+     remove_folder (rio_dev, folder_num, card_number);
    }
    else 
    {
@@ -377,7 +372,7 @@ main(int argc, char *argv[])
              exit (0);
            }
        }
-     remove_song (rio_dev, song_num, folder_num);
+     remove_song (rio_dev, song_num, folder_num, card_number);
     }
 
    /* Close device */
@@ -394,9 +389,9 @@ main(int argc, char *argv[])
 
 int
 #ifndef WITH_USBDEVFS
-remove_folder (int rio_dev, int folder_num)
+remove_folder (int rio_dev, int folder_num, int card_number)
 #else
-remove_folder (struct usbdevice *rio_dev, int folder_num)
+remove_folder (struct usbdevice *rio_dev, int folder_num, int card_number)
 #endif
 {
   folder_entry     *folder;
@@ -404,11 +399,11 @@ remove_folder (struct usbdevice *rio_dev, int folder_num)
   GList            *folders, *songs;
 
   /* Read folder & song block */
-  folders = read_folder_entries (rio_dev);
+  folders = read_folder_entries (rio_dev,card_number);
   if ( folder_num > g_list_length (folders)-1 )
     return -1;
   else
-    songs   = read_song_entries (rio_dev, folders, folder_num); 
+    songs   = read_song_entries (rio_dev, folders, folder_num, card_number); 
 
 
   folder  = (folder_entry *)g_list_nth_data (g_list_first(folders), folder_num);
@@ -417,12 +412,12 @@ remove_folder (struct usbdevice *rio_dev, int folder_num)
   printf ("Removing folder <%s>...\n", folder->name1);
 
   for (song_num = g_list_length (songs) - 1; song_num >= 0; song_num--)
-    remove_song (rio_dev, song_num, folder_num);
+    remove_song (rio_dev, song_num, folder_num, card_number);
 
   folders = g_list_remove (folders, (gpointer) folder);
 
-  send_command (rio_dev, 0x4c, folder_num, 0xff);
-  write_folder_entries (rio_dev, folders);
+  send_command (rio_dev, 0x4c, ((folder_num << 8) | 0xff), card_number);
+  write_folder_entries (rio_dev, folders,card_number);
   send_command (rio_dev, 0x42, 0, 0);
   send_command (rio_dev, 0x42, 0, 0);
   folder_block_offset = send_command (rio_dev, 0x43, 0x0, 0x0);
@@ -432,18 +427,18 @@ remove_folder (struct usbdevice *rio_dev, int folder_num)
 #endif
 
   /* Tell Rio where the root folder block is. */
-  send_folder_location (rio_dev, folder_block_offset, 0);
+  send_folder_location (rio_dev, folder_block_offset, 0, card_number);
 
   /* Not really sure what this does */
-  send_command (rio_dev, 0x58, 0x0, 0x0);
+  send_command (rio_dev, 0x58, 0x0, card_number);
 
   return 0;
 }
 
 #ifndef WITH_USBDEVFS
-int remove_song (int rio_dev, int song_num, int folder_num)
+int remove_song (int rio_dev, int song_num, int folder_num,int card_number)
 #else
-int remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
+int remove_song (struct usbdevice *rio_dev, int song_num, int folder_num, int card_number)
 #endif
 {
   song_entry       *song;
@@ -452,13 +447,13 @@ int remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
   GList            *folders, *songs;
 
    /* Read folder & song block */
-   folders = read_folder_entries (rio_dev);
+   folders = read_folder_entries (rio_dev, card_number);
    if ( folder_num > g_list_length (folders)-1 )
    {
-     songs   = read_song_entries (rio_dev, folders, 0); /* use folder 0 by default */
+     songs   = read_song_entries (rio_dev, folders, 0, card_number); /* use folder 0 by default */
      folder_num = 0;
    } else
-     songs   = read_song_entries (rio_dev, folders, folder_num); 
+     songs   = read_song_entries (rio_dev, folders, folder_num, card_number); 
 
    /* Remove the song entry from the list */
    if ( song_num > g_list_length (songs) -1)
@@ -472,10 +467,10 @@ int remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
    songs = g_list_remove (songs, (gpointer)song);
 
    /* Send remove command to the rio */
-   send_command (rio_dev, 0x4c, ((folder_num << 8) | song_num), 0);
+   send_command (rio_dev, 0x4c, ((folder_num << 8) | song_num), card_number);
 
    /* Write song block to the correct folder */
-   write_song_entries (rio_dev, folder_num, songs );
+   write_song_entries (rio_dev, folder_num, songs,card_number );
    send_command (rio_dev, 0x42, 0, 0);
    send_command (rio_dev, 0x42, 0, 0);
 
@@ -490,7 +485,7 @@ int remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
    f_entry->offset = song_block_offset;
    f_entry->fst_free_entry_off -= 0x800;
 
-   write_folder_entries ( rio_dev, folders );
+   write_folder_entries ( rio_dev, folders,card_number );
    send_command (rio_dev, 0x42, 0, 0);
    send_command (rio_dev, 0x42, 0, 0);
    folder_block_offset = send_command (rio_dev, 0x43, 0x0, 0x0);
@@ -500,18 +495,19 @@ int remove_song (struct usbdevice *rio_dev, int song_num, int folder_num)
 #endif
 
    /* Tell Rio where the root folder block is. */
-   send_folder_location (rio_dev, folder_block_offset, folder_num);
+   send_folder_location (rio_dev, folder_block_offset, folder_num,card_number);
 
    /* Not really sure what this does */
-   send_command (rio_dev, 0x58, 0x0, 0x0);
+   send_command (rio_dev, 0x58, 0x0, card_number);
 
    return 0;
 }
 
-static char const shortopts[] = "F:abwhv";
+static char const shortopts[] = "F:xabwhv";
 static struct option const longopts[] =
 {
   {"folder", required_argument, NULL, 'F'},
+  {"external", no_argument, NULL, 'x'},
   {"automatic", no_argument, NULL, 'a'},
   {"byname",no_argument, NULL, 'b'},
   {"wholefolder", no_argument, NULL, 'w'},
@@ -525,6 +521,7 @@ static char const *const option_help[] =
 "Input options:",
 "",
 "  -F x      --folder x         Delete song(s) from folder of index x",
+"  -x        --external         Delete from external memory card",
 "  -a        --automatic        Deletes without prompting",
 "  -w        --wholefolder      Treats arguments as folder names/indices",
 "                               Will delete entire folder at once",
@@ -545,7 +542,7 @@ static char const *const option_help[] =
 /* Process switches and filenames.  */
 
 void
-get_some_switches (int argc, char *argv[], int *folder_num, int *automat)
+get_some_switches (int argc, char *argv[], int *folder_num, int *automat, int *card_number)
 {
     register int optc;
     char const * const *p;
@@ -556,6 +553,10 @@ get_some_switches (int argc, char *argv[], int *folder_num, int *automat)
     while ((optc = getopt_long (argc, argv, shortopts, longopts, (int *) 0))
            != -1) {
          switch (optc) {
+	    case 'x':
+		*card_number=1;
+		break;
+
             case 'F':
                 folder_num_set = 1;
 		folder_num_string=malloc(sizeof(optarg));
